# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

from math import *
import math
import bpy
import bmesh
import rna_keymap_ui

bl_info = {
    "name": "Loop Ring (LR) Select",
    "description": "Grow, shrink, select Loops and Rings quickly",
    "author": "Bartosz Styperek",
    "version": (0, 1, 2),
    "blender": (2, 93, 0),
    "location": "L-key pie menu",
    "warning": "",
    "wiki_url": "",
    "category": "Mesh"}


def other_not_selected_edges(vert, edge):
    for ed in vert.link_edges:
        if ed != edge and not ed.select:
            yield ed


def other_edges(vert, edge):
    for ed in vert.link_edges:
        if ed != edge:
            yield ed


def edges_angle(edgeA, edgeB, adj_vert):
    vec_a = edgeA.other_vert(adj_vert).co - adj_vert.co   # from A -> Vert
    vec_b = edgeB.other_vert(adj_vert).co - adj_vert.co  # from vert  -> B
    return vec_a.angle(vec_b)


CUT_ANGLE = math.pi/2
def get_min_angle_edge(e, vert):
    best_angle = CUT_ANGLE  # if angle more than 90 + X deg threshold
    best_edge = None
    for other_edge in other_edges(vert, e):
        # angle = math.acos(edges_angle(e, other_edge, vert))
        angle = edges_angle(e, other_edge, vert)
        if angle > best_angle:  # just to check if 2 edges that passed were to similar
            best_angle = angle
            best_edge = other_edge
    return best_edge



def get_next_edge_loop(current_vert, current_edge):
    e_link_faces = set(current_edge.link_faces)
    vert_link_edges_count = len(current_vert.link_edges)
    vert_link_faces_count = len(current_vert.link_faces)
    new_edge = None
    if vert_link_edges_count == 4:
        if vert_link_faces_count == 4:  # use linked faces to predict next edge
            for other_edge in other_edges(current_vert, current_edge):
                if not e_link_faces & set(other_edge.link_faces):
                    new_edge = other_edge
                    break
        elif vert_link_faces_count == 3:  # es edge.loops to predict next edge
            for e_loop in current_edge.link_loops:
                next_e = e_loop.link_loop_prev.link_loop_radial_next.link_loop_prev.edge if e_loop.vert == current_vert else e_loop.link_loop_next.link_loop_radial_next.link_loop_next.edge
                if next_e in current_vert.link_edges:
                    new_edge = next_e
                    break
        else:  # try using dot product to predict next edge
            new_edge = get_min_angle_edge(current_edge, current_vert)

    elif vert_link_edges_count == 3:  # if vert.linked_e != 4
        if vert_link_faces_count == 3:  # Y split, so just pick one
            new_edge = get_min_angle_edge(current_edge, current_vert)
        elif vert_link_faces_count == 2:
            if len(current_edge.link_loops) == 2:  # 2 faces adjacent to curre edge - reached border T, break
                pass
            else:  # 1 faces adjacent to curre edge - 1 face adj to next edge.
                for other_edge in other_edges(current_vert, current_edge):
                    if not e_link_faces & set(other_edge.link_faces) and not other_edge.select:  # just avoid edge that has common face with current_e
                        new_edge = other_edge
                        break
        else:  # non manifold mesh it seems. Just try picking min edge edge
            new_edge = get_min_angle_edge(current_edge, current_vert)
    elif vert_link_edges_count == 2:  # straing line?
        new_edge = get_min_angle_edge(current_edge, current_vert)
    else:  # more than 4 edges, or just one edge()
        new_edge = get_min_angle_edge(current_edge, current_vert)

    return new_edge


class MESH_OT_SelectLoops(bpy.types.Operator):
    bl_idname = "mesh.select_loops"
    bl_label = "Select Loop"
    bl_description = "Select loop"
    bl_options = {"REGISTER", "UNDO"}

    max_angle: bpy.props.FloatProperty(name='Search Angle', description='Bigger angle - search broader cone for next edge',
                    default=math.pi/4, min=0, max=math.pi/2, unit='ROTATION', step=20, precision=0)  # , subtype='ANGLE'

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        global CUT_ANGLE
        CUT_ANGLE = math.pi/2 + (math.pi/2 - self.max_angle)

        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_edges = [e for e in bm.edges if e.select]
        edges_to_select_loops = [[e] for e in bm.edges if e.select]
        for e_ith, e in enumerate(selected_edges):
            cut_l = cut_r = 0  # do we cut out right side to  [..,e,..] or left
            go_right = 1
            for e_vert in e.verts:  # 2 edge_verts
                prev_count = 0  # how many selected edges there were in prev step
                step = 0  # how many steps we took to left(right) from e
                current_edge = e
                current_vert = e_vert
                break_while = False
                while len(edges_to_select_loops[e_ith]) != prev_count and current_edge and not break_while:  # if no new edges added exit loop
                    break_while = False
                    prev_count = len(edges_to_select_loops[e_ith])
                    new_edge = get_next_edge_loop(current_vert, current_edge)

                    #* Clean selected list of loops, by removing unecessary sides.
                    if new_edge:
                        #! avoid looping indefinietly sometimes.. corner case
                        if (go_right and new_edge in edges_to_select_loops[e_ith][-step:]) or (not go_right and new_edge in edges_to_select_loops[e_ith][:step]):
                            break_while = True
                            break # break while

                        if go_right:
                            edges_to_select_loops[e_ith].append(new_edge)
                        else:
                            edges_to_select_loops[e_ith].insert(0, new_edge)
                        step += 1
                        if new_edge.select:  # break if we hit already selected edge...
                            if go_right:
                                cut_r = step
                            else:
                                cut_l = step
                            break_while = True
                        current_edge = new_edge
                        current_vert = current_edge.other_vert(current_vert)
                    else:
                        current_edge = None

                go_right = 0  # now go left from edge e
                # * while end
            if not cut_l and not cut_r:
                pass
            elif cut_l and not cut_r:  # remove right part from e [..,e,..]
                if cut_l == 1:  # if cut on neibor edge - just ignore that
                    continue
                edges_to_select_loops[e_ith] = edges_to_select_loops[e_ith][:cut_l+1]  # remove right part
            elif not cut_l and cut_r:  # remove left part from e [..,e,..]
                if cut_r == 1:  # if cut on neibor edge - just ignore that
                    continue
                edges_to_select_loops[e_ith] = edges_to_select_loops[e_ith][-(cut_r+1):]  # remove left part
            elif cut_l and cut_r:
                if cut_l == cut_r:
                    edges_to_select_loops[e_ith] = edges_to_select_loops[e_ith][:cut_l]
                else:
                    use_right = True if cut_r < cut_l else False
                    edges_to_select_loops[e_ith] = edges_to_select_loops[e_ith][-(cut_r+1):] if use_right else edges_to_select_loops[e_ith][:cut_l+1]

        for eloops in edges_to_select_loops:
            for e in eloops:  # may contain duplicates but whatever
                e.select = True

        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


class MESH_OT_GrowLoop(bpy.types.Operator):
    bl_idname = "mesh.grow_loop"
    bl_label = "Grow Loop"
    bl_description = "Grow loop"
    bl_options = {"REGISTER", "UNDO"}

    max_angle: bpy.props.FloatProperty(name='Search Angle', description='Bigger angle - search broader cone for next edge',
                                       default=math.pi/4, min=0, max=math.pi/2, unit='ROTATION', step=20, precision=0)  # , subtype='ANGLE'
    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        global CUT_ANGLE
        CUT_ANGLE = math.pi/2 + (math.pi/2 - self.max_angle)
        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_edges = [e for e in bm.edges if e.select]
        edges_to_select = []
        for current_edge in selected_edges:
            for current_vert in current_edge.verts:
                new_edge = get_next_edge_loop(current_vert, current_edge)
                if new_edge:
                    edges_to_select.append(new_edge)

        for e in edges_to_select:  # may contain duplicates but whatever
            e.select = True

        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


class MESH_OT_ShrinkLoop(bpy.types.Operator):
    bl_idname = "mesh.shrink_loop"
    bl_label = "Shrink Loop"
    bl_description = "Shrink loop"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_edges = [e for e in bm.edges if e.select]
        deselect_border_edges = []
        for e in selected_edges:
            e_link_faces = set(e.link_faces)
            for vert in e.verts:
                other_nonsel_e = list(other_not_selected_edges(vert, e))
                other_nonsel_e_count = len(other_nonsel_e)
                if len(vert.link_edges) == 4 and other_nonsel_e_count == 3:  # vert is the end of loop. So mark it for deselect
                    deselect_border_edges.append(e)
                elif len(vert.link_edges) == 2 and other_nonsel_e_count == 1:
                    deselect_border_edges.append(e)

                elif len(vert.link_edges) == 3 and other_nonsel_e_count == 2:
                    deselect_border_edges.append(e)
        for e in deselect_border_edges:  # may contain duplicates but whatever
            e.select = False
        bm.select_history.validate()
        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


class MESH_OT_LoopToRing(bpy.types.Operator):
    bl_idname = "mesh.loops_to_rigns"
    bl_label = "Loops to rings"
    bl_description = "Loops to rings"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_faces = [f for f in bm.faces if all([v.select for v in f.verts])]
        edges_to_select = []
        for f in selected_faces:
            for e in f.edges:
                if not e.select:
                    edges_to_select.append(e)
        for e in bm.edges:
            e.select = False
        e = None
        for e in set(edges_to_select):
            e.select = True
        bm.select_history.validate()
        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


class MESH_OT_GrowRing(bpy.types.Operator):
    bl_idname = "mesh.grow_ring"
    bl_label = "Grow Ring"
    bl_description = "Grow ring"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_edges = [e for e in bm.edges if e.select]
        edges_to_select = []
        for e in selected_edges:
            for loop in e.link_loops:
                if len(loop.face.edges) == 4:
                    edges_to_select.append(loop.link_loop_next.link_loop_next.edge)
        for e in edges_to_select:
            e.select = True
        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


class MESH_OT_SelectRing(bpy.types.Operator):
    bl_idname = "mesh.select_rings"
    bl_label = "Select Ring"
    bl_description = "Select ring"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_edges = [e for e in bm.edges if e.select]
        edges_to_select_rings = [[e] for e in bm.edges if e.select]
        for e_ith, e in enumerate(selected_edges):
            go_right = 1
            cut_l = cut_r = 0  # do we cut out right side to  [..,e,..] or left
            for loop in e.link_loops:  # go left and right from edge
                prev_count = 0  # how many selected edges there were in preve iteration
                step = 0  # how many steps we took to left(right) from e
                next_loop = loop
                while len(edges_to_select_rings[e_ith]) != prev_count and next_loop:  # if no new edges added exit loop
                    prev_count = len(edges_to_select_rings[e_ith])
                    if len(next_loop.face.edges) == 4:
                        current_edge = next_loop.link_loop_next.link_loop_next.edge
                        if go_right:
                            edges_to_select_rings[e_ith].append(current_edge)
                        else:
                            edges_to_select_rings[e_ith].insert(0, current_edge)
                        step += 1
                        if current_edge.select:  # break if we hit already selected edge...
                            if go_right:
                                cut_r = step
                            else:
                                cut_l = step
                            break
                        next_loop = next_loop.link_loop_next.link_loop_next.link_loop_radial_next if not edges_to_select_rings[e_ith][-1*go_right].is_boundary else False

                    else:  # non qud polyogn - stop going
                        break
                go_right = 0  # now go left from edge e

            if not cut_l and not cut_r:
                pass
            elif cut_l and not cut_r:  # remove right part from e [..,e,..]
                if cut_l == 1: #if cut on neibor edge - just ignore that
                    continue
                edges_to_select_rings[e_ith] = edges_to_select_rings[e_ith][:cut_l+1]  # remove right part
            elif not cut_l and cut_r:  # remove left part from e [..,e,..]
                if cut_r == 1:  # if cut on neibor edge - just ignore that
                    continue
                edges_to_select_rings[e_ith] = edges_to_select_rings[e_ith][-(cut_r+1):]  # remove left part
            elif cut_l and cut_r:
                if cut_l == cut_r:
                    edges_to_select_rings[e_ith] = edges_to_select_rings[e_ith][:cut_l]
                else:
                    use_right = True if cut_r < cut_l else False
                    edges_to_select_rings[e_ith] = edges_to_select_rings[e_ith][-(cut_r+1):] if use_right else edges_to_select_rings[e_ith][:cut_l+1]

        for e_ring in edges_to_select_rings:
            for e in e_ring:
                e.select = True
        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


class MESH_OT_ShrinkRing(bpy.types.Operator):
    bl_idname = "mesh.shrink_ring"
    bl_label = "Shrink Ring"
    bl_description = "Shrink ring"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def execute(self, context):
        obj = context.active_object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        selected_edges = [e for e in bm.edges if e.select]
        edges_to_deselect = []
        for e in selected_edges:
            for loop in e.link_loops:  # ? two loops if edge has 2 faces. What if not....
                if len(loop.face.edges) == 4 and not loop.link_loop_next.link_loop_next.edge.select:
                    edges_to_deselect.append(e)
                    break
                if len(loop.face.edges) != 4:
                    edges_to_deselect.append(e)
                    break
        e = None
        for e in edges_to_deselect:
            e.select = False
        bm.select_history.validate()
        bmesh.update_edit_mesh(me)
        return {"FINISHED"}


def my_get_sorted_loops2(self, bm):
    remaining_edges_ids = [e.index for e in bm.edges if e.select]
    edge_strips = []
    vert_strips = []

    while remaining_edges_ids:
        current_edge_loop = []
        current_vert_loop = []
        current_edge = bm.edges[remaining_edges_ids.pop()]
        current_edge_loop.append(current_edge)

        def follow_edge_loop(curr_edge, old_vert, add_right):
            while True:  # go vert0 way, till no lined edges (line stops)
                current_vert = curr_edge.other_vert(old_vert)
                next_edges = [link_e for link_e in current_vert.link_edges if link_e.select and link_e != curr_edge]
                if add_right:
                    current_vert_loop.append(current_vert)
                else:
                    current_vert_loop.insert(0, current_vert)
                if next_edges and next_edges[0] not in current_edge_loop:  # just ignore if there are more linked edges selected. Maybe give error - bad selection
                    if add_right:
                        current_edge_loop.append(next_edges[0])
                        # current_vert_loop.append(current_vert)
                    else:
                        current_edge_loop.insert(0, next_edges[0])
                        # current_vert_loop.insert(0,current_vert)
                    remaining_edges_ids.remove(next_edges[0].index)
                    curr_edge = next_edges[0]
                    old_vert = current_vert  # right vert becomes left
                else:
                    break
        follow_edge_loop(current_edge, current_edge.verts[0], True)  # go in direction of vert 1
        follow_edge_loop(current_edge, current_edge.verts[1], False)  # go in direction of vert 0

        edge_strips.append(current_edge_loop)
        vert_strips.append(current_vert_loop)

    return vert_strips, edge_strips


class ST_MT_Menu(bpy.types.Menu):
    bl_idname = "ST_MT_SelectToolsMenu"
    bl_label = "Select Tools"

    draw_as_box = True

    @classmethod
    def poll(cls, context):
        return context.mode == 'EDIT_MESH'

    def draw(self, context):
        layout = self.layout
        pie = layout.menu_pie()
        # 1 - LEFT
        box = pie.box()
        col = box.column(align=True)
        col.operator("mesh.select_rings")
        col.operator("mesh.grow_ring")
        col.operator("mesh.shrink_ring")

        # 2 - RIGHT
        box = pie.box()
        col = box.column(align=True)
        col.operator("mesh.select_loops")
        col.operator("mesh.grow_loop")
        col.operator("mesh.shrink_loop")

        # 3 - BOTTOM
        box = pie.box()

        col = box.column()
        # subrow = col.row(align=True)
        col.operator("mesh.loops_to_rigns")
        # col.operator("mesh.faces_select_linked_flat")

        # 4 - TOP
        box = pie.box()
        col = box.column(align=True)
        col.operator("mesh.loop_to_region")
        col.operator("mesh.region_to_loop")

        # 5 - TOP - LEFT
        pie.separator()
        # 6 - TOP - RIGHT
        pie.separator()
        # 7 - BOTTOM - LEFT
        pie.separator()
        # 8 - BOTTOM - RIGHT
        pie.separator()


class SelectToolsPreferences(bpy.types.AddonPreferences):
    # this must match the addon name, use '__package__'
    # when defining this in a submodule of a python package.
    bl_idname = __name__

    # category: bpy.props.StringProperty(name="Tab Category", description="Choose a name for the category of the panel", default="Tools", update=update_panel)

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        split = box.split()
        col = split.column()
        col.label(text='Hotkey')
        col.separator()
        wm = bpy.context.window_manager
        kc = wm.keyconfigs.user  # cos we override addon hotkey,  with this
        km = kc.keymaps['Mesh']
        kmi = get_hotkey_entry_item(km, 'wm.call_menu_pie', "ST_MT_SelectToolsMenu")
        if kmi:
            col.context_pointer_set("keymap", km)
            rna_keymap_ui.draw_kmi([], kc, km, kmi, col, 0)
        else:
            col.label(text="No hotkey entry found")
            col.operator(ST_OT_ReBevel_Add_Hotkey.bl_idname, text="Add hotkey entry", icon='PLUS')


def get_hotkey_entry_item(km, kmi_name, kmi_value):
    '''
    returns hotkey of specific type, with specific properties.name (keymap is not a dict, so referencing by keys is not enough
    if there are multiple hotkeys!)
    '''
    for i, km_item in enumerate(km.keymap_items):
        if km.keymap_items.keys()[i] == kmi_name:
            if km.keymap_items[i].properties.name == kmi_value:
                return km_item
    return None


class ST_OT_SelTool_Add_Hotkey(bpy.types.Operator):
    ''' Add hotkey entry '''
    bl_idname = "selecttool.add_hotkey"
    bl_label = "SelectTool Init Hotkey"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):
        register_keymap()
        self.report({'INFO'}, "Hotkey added in User Preferences -> Input -> 3D view")
        return {'FINISHED'}


addon_keymaps = []


def register_keymap():
    '''AFAIK this is overrriden later by user set hotkey in keyconfigs.user.
    That is why this wont always reset do defautl hotkey'''
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon

    km = kc.keymaps.new(name='Mesh', space_type='EMPTY')
    kmi = km.keymap_items.new('wm.call_menu_pie', 'L', 'PRESS', shift=False)
    kmi.properties.name = "ST_MT_SelectToolsMenu"
    addon_keymaps.append((km, kmi))


def unregister_keymap():
    wm = bpy.context.window_manager
    for km, kmi in addon_keymaps:
        if len(km.keymap_items) > 0:  # why this is empty/broken of F8 scripts.reload?
            km.keymap_items.remove(kmi)
        # if km in wm.keyconfigs.addon.keymaps.values():
        #     wm.keyconfigs.addon.keymaps.remove(km)
    addon_keymaps.clear()


classes = (
    MESH_OT_GrowLoop,
    MESH_OT_GrowRing,
    MESH_OT_SelectLoops,
    MESH_OT_SelectRing,
    MESH_OT_LoopToRing,
    MESH_OT_ShrinkLoop,
    MESH_OT_ShrinkRing,
    ST_OT_SelTool_Add_Hotkey,
    ST_MT_Menu,
    SelectToolsPreferences,
)


def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
    register_keymap()


def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)
    unregister_keymap()
